@@ -8,9 +8,11 @@ use utf8;
 use App::ArduinoBuilder::Builder 'build_archive', 'build_object_files', 'link_executable', 'run_hook';
 use App::ArduinoBuilder::CommandRunner;
 use App::ArduinoBuilder::Config 'get_os_name';
+use App::ArduinoBuilder::Discovery;
 use App::ArduinoBuilder::FilePath 'find_latest_revision_dir', 'list_sub_directories', 'find_all_files_with_extensions';
 use App::ArduinoBuilder::Logger;
-use App::ArduinoBuilder::System 'find_arduino_dir', 'system_cwd';
+use App::ArduinoBuilder::Monitor;
+use App::ArduinoBuilder::System 'find_arduino_dir', 'system_cwd', 'execute_cmd';
 
 use File::Basename;
 use File::Path 'remove_tree';
@@ -19,19 +21,19 @@ use Getopt::Long;
 use List::Util 'any', 'none', 'first';
 use Pod::Usage;
 
-our $VERSION = '0.05';
+our $VERSION = '0.07';
+
+# User agent used for the pluggable discovery and pluggable monitor tools.
+our $TOOLS_USER_AGENT = "\"App::ArduinoBuilder ${VERSION}\"";
 
 sub Run {
-  my $project_dir;
-  my $build_dir;
-
   my $config = App::ArduinoBuilder::Config->new();
 
   my (@skip, @force, @only);
   GetOptions(
       'help|h' => sub { pod2usage(-exitval => 0, -verbose => 2)},
-      'project-dir|project|p=s' => \$project_dir,
-      'build-dir|build|b=s' => \$build_dir,
+      'project-dir|project|p=s' => sub { $config->set('builder.project_dir' => $_[1], allow_override => 1) },
+      'build-dir|build|b=s' => sub { $config->set('builder.internal.build_dir' => $_[1], allow_override => 1) },
       'log-level|l=s' => sub { App::ArduinoBuilder::Logger::set_log_level($_[1]) },
       'config|c=s%' => sub { $config->set($_[1] => $_[2], allow_override => 1) },
       'menu=s%' => sub { $config->set('builder.menu.'.$_[1] => $_[2], allow_override => 1) },
@@ -39,63 +41,70 @@ sub Run {
       'force=s@' => sub { push @force, split /,/, $_[1] },  # even if it would be skipped by the dependency checker
       'only=s@' => sub { push @only, split /,/, $_[1] },  # run only these steps (skip all others)
       'stack-trace-on-error|stack' => sub { App::ArduinoBuilder::Logger::print_stack_on_fatal_error(1) },
-      'j=i' => sub { $config->set('builder.parallelize' => $_[1], allow_override => 1) },
+      'parallelize|j=i' => sub { $config->set('builder.parallelize' => $_[1], allow_override => 1) },
+      'target-port|port=s' => sub { $config->append('builder.upload.port' => $_[1], ',')},
     ) or pod2usage(-exitval => 2, -verbose =>0);
 
-  fatal "More than one command specified: ".join(' ', @ARGV) if @ARGV > 1;
-  my $command = @ARGV ? $ARGV[0] : 'build';
+  if (my @unknown = grep { !/^(clean|build|discover|upload|monitor)$/ } @ARGV) {
+    fatal "Unknown command%s: %s", (@unknown > 1 ? 's' : ''), join(', ', @unknown);
+  }
+
+  generate_project_config($config);
+
+  push @ARGV, 'build' unless @ARGV;
+
+  if (grep { /^clean$/ } @ARGV) {
+    clean($config);
+  }
+  if (grep { /^build$/ } @ARGV) {
+    build($config, \@skip, \@force, \@only);
+  }
+  if (grep { /^discover$/ } @ARGV && ! grep { /^(upload|monitor)$/ } @ARGV) {
+    discover($config);
+  }
+  if (grep { /^upload$/ } @ARGV) {
+    discover($config);
+    upload($config);
+  }
+  if (grep { /^monitor$/ } @ARGV) {
+    # The upload process potentially modifies the board port, so we run the
+    # discovery here even if it was run on upload.
+    discover($config);
+    monitor($config);
+  }
+}
+
+sub generate_project_config {
+  my ($config) = @_;
 
   my $project_dir_is_cwd = 0;
-  if (!$project_dir) {
+  my $project_dir;
+  if (!$config->exists('builder.project_dir')) {
     $project_dir_is_cwd = 1;
     $project_dir = system_cwd();
+    $config->set('builder.project_dir' => $project_dir);
+  } else {
+    $project_dir = $config->get('builder.project_dir');
   }
 
   $config->read_file(catfile($project_dir, 'arduino_builder.local'), allow_missing => 1);
   $config->read_file(catfile($project_dir, 'arduino_builder.config'), allow_missing => 1);
 
-  $config->set('builder.project_dir' => $project_dir);
-
-  if (!$build_dir) {
+  my $build_dir;
+  if (!$config->exists('builder.internal.build_dir')) {
     if ($config->exists('builder.default_build_dir')) {
       $build_dir = $config->get('builder.default_build_dir');
+      $config->set('builder.internal.build_dir_from_default' => 1);
     } elsif (!$project_dir_is_cwd) {
       $build_dir = system_cwd();
     } else {
       fatal 'No builder.default_build_dir config and --build_dir was not passed when building from the project directory.';
     }
+    $config->set('builder.internal.build_dir' => $build_dir);
   }
-
-  if ($command eq 'build') {
-    build($config, $build_dir, \@skip, \@force, \@only);
-  } elsif ($command eq 'clean') {
-    clean($config, $build_dir);
-  } else {
-    fatal "Unknown command: ${command}";
-  }
-}
-
-sub clean {
-  my ($config, $build_dir) = @_;
-
-  # TODO: add a way to clean only parts of the projects.
-  if ($build_dir eq $config->get('builder.default_build_dir')) {
-    remove_tree($build_dir, {safe => 1, keep_root => 1});
-  } else {
-    warning "For safety reason we can only clean build directory specified in the project";
-    warning "config. You should run `rm -rf` manually.";
-    fatal "Not cleaning context dependent directory: ${build_dir}";
-  }
-}
-
-sub build {
-  my ($config, $build_dir, @array_args) = @_;
-  my @skip = @{$array_args[0]};
-  my @force = @{$array_args[1]};
-  my @only = @{$array_args[2]};
+  $config->set('build.path' => $config->get('builder.internal.build_dir'));
 
   if (!$config->exists('builder.source.path')) {
-    my $project_dir = $config->get('builder.project_dir');
     my $d = first { -d catdir($project_dir, $_) } qw(src srcs source sources);
     if (defined $d) {
       $config->set('builder.source.path' => catdir($project_dir, $d));
@@ -108,11 +117,17 @@ sub build {
     $config->set('builder.source.is_recursive' => 1, ignore_existing => 1);
   }
 
+  my $arduino_dir;
+  if ($config->exists('builder.arduino.install_dir')) {
+    $arduino_dir = $config->get('builder.arduino.install_dir');
+  } else {
+    $arduino_dir = find_arduino_dir();
+  }
+
   if (!$config->exists('builder.package.path')) {
     fatal 'At least one of builder.package.path or builder.package.name must be specified in the config' unless $config->exists('builder.package.name');
     # TODO: the core package can also be installed in a "hardware" directory in
     # the sketch directory. We should search for it there.
-    my $arduino_dir = find_arduino_dir();
     fatal "The builder.package.path config is not set and Arduino installation directory not found" unless $arduino_dir;
     debug "Using arduino directory: ${arduino_dir}";
     my $package_name = $config->get('builder.package.name');
@@ -193,7 +208,6 @@ sub build {
   # todo: name, _id, build.fqbn, and the time options
   $config->set('build.source.path' => $config->get('builder.source.path'));
   $config->set('sketch_path' => $config->get('builder.source.path'));
-  $config->set('build.path' => $build_dir);
   $config->set('build.project_name' => $config->get('builder.project_name'));
   $config->set('build.arch' => uc($config->get('builder.package.arch')));  # Undocumented but it seems that it’s always upper case.
   $config->set('build.core.path', catdir($hardware_path, 'cores', $config->get('build.core')));
@@ -201,6 +215,13 @@ sub build {
   $config->set('build.variant.path', catdir($hardware_path, 'variants', $config->get('build.variant')));
 
   my @tools_dirs = (catdir($package_path, 'tools'));
+  if ($arduino_dir) {
+    push @tools_dirs, catdir($arduino_dir, 'packages', 'builtin', 'tools');
+    push @tools_dirs, catdir($arduino_dir, 'packages', 'arduino', 'tools');
+  } else {
+    warning 'The Arduino GUI directory could not be found, we won’t use the builtin tools.';
+  }
+
   for my $tools_dir (@tools_dirs) {
     next unless -d $tools_dir;
     my @tools = list_sub_directories($tools_dir);
@@ -208,9 +229,11 @@ sub build {
       debug "Found tool: $t";
       my $tool_path = catdir($tools_dir, $t);
       my $latest_tool_path = find_latest_revision_dir($tool_path);
-      $config->set("runtime.tools.${t}.path", $latest_tool_path);
+      # That one could point to the latest version found across all packages
+      # instead of the first version (possibly that of "our" package).
+      $config->set("runtime.tools.${t}.path", $latest_tool_path, ignore_existing => 1);
       for my $v (list_sub_directories($tools_dir)) {
-        $config->set("runtime.tools.${t}-${v}.path", catdir($tool_path, $v));
+        $config->set("runtime.tools.${t}-${v}.path", catdir($tool_path, $v), ignore_existing => 1);
       }
     }
   }
@@ -231,6 +254,29 @@ sub build {
     default_runner()->set_max_parallel_tasks($config->get('builder.parallelize'));
   }
 
+  return 1;
+}
+
+sub clean {
+  my ($config) = @_;
+
+  # TODO: add a way to clean only parts of the projects.
+  my $build_dir = $config->get('builder.internal.build_dir');
+  if ($config->get('builder.internal.build_dir_from_default', default => 0)) {
+    remove_tree($build_dir, {safe => 1, keep_root => 1});
+  } else {
+    warning "For safety reason we can only clean build directory specified in the project";
+    warning "config. You should run `rm -rf` manually.";
+    fatal "Not cleaning context dependent directory: ${build_dir}";
+  }
+}
+
+sub build {
+  my ($config, @array_args) = @_;
+  my @skip = @{$array_args[0]};
+  my @force = @{$array_args[1]};
+  my @only = @{$array_args[2]};
+
   my $run_step = sub {
     my ($step) = @_;
     return (none { $_ eq $step } @skip) && (!@only || any { $_ eq $step } @only);
@@ -256,7 +302,7 @@ sub build {
   if ($run_step->('core')) {
     info 'Building core...';
     $builder->run_hook('core.prebuild');
-    my $built_core = $builder->build_archive([$config->get('build.core.path'), $config->get('build.variant.path')], catdir($build_dir, 'core'), 'core.a', $force->('core'));
+    my $built_core = $builder->build_archive([$config->get('build.core.path'), $config->get('build.variant.path')], catdir($config->get('build.path'), 'core'), 'core.a', $force->('core'));
     info ($built_core ? '  Success' : '  Already up-to-date');
     $built_something |= $built_core;
     $builder->run_hook('core.postbuild');
@@ -292,12 +338,13 @@ sub build {
   # and also the precompiled
   if ($run_step->('libraries')) {
     info 'Building libraries...';
+    $builder->run_hook('libraries.prebuild');
     for my $l (@all_libs) {
       # Todo: we could add "lib-$l" pseudo-steps, but we need to take care of the --only
       # interaction with the 'libraries' step.
       info '  Building library %s...', $lib_config->get("${l}.name");
       my $base_dir = $lib_config->get($l);
-      my $output_dir = catdir($build_dir, 'libs', $l);
+      my $output_dir = catdir($config->get('build.path'), 'libs', $l);
       my $built_lib;
       if ($lib_config->get("${l}.is_flat")) {
         my $utility_dir = catdir($base_dir, 'utility');
@@ -309,6 +356,7 @@ sub build {
       info ($built_lib ? '    Success' : '    Already up-to-date');
       $built_something |= $built_lib;
     }
+    $builder->run_hook('libraries.postbuild');
   }
 
   $config->append('includes', '"-I'.$config->get('builder.source.path').'"');
@@ -319,7 +367,7 @@ sub build {
     # TODO: add configuration option for the ignored directories and also a way to
     # build only the code inside the src/ directory
     my $built_sketch = $builder->build_object_files(
-        $config->get('builder.source.path'), catdir($build_dir, 'sketch'),
+        $config->get('builder.source.path'), catdir($config->get('build.path'), 'sketch'),
         [], $force->('sketch'), $config->get('builder.source.is_recursive'));
     info ($built_sketch ? '  Success' : '  Already up-to-date');
     $built_something |= $built_sketch;
@@ -328,8 +376,10 @@ sub build {
   # Bug: there is a similar bug to the one in build_archive: if a source file is
   # removed, we won’t remove it’s object file. I guess we could try to detect it.
   # Meanwhile it’s probably acceptable to ask for a cleanup from time to time.
-  my @object_files = find_all_files_with_extensions(catdir($build_dir, 'sketch'), ['o']);
-  push @object_files, find_all_files_with_extensions(catdir($build_dir, 'libs'), ['o']);
+  my @object_files = find_all_files_with_extensions(catdir($config->get('build.path'), 'sketch'), ['o']);
+  for my $l (@all_libs) {
+    push @object_files, find_all_files_with_extensions(catdir($config->get('build.path'), 'libs', $l), ['o']);
+  }
   debug 'Object files: '.join(', ', @object_files);
 
   info 'Linking binary...';
@@ -353,6 +403,94 @@ sub build {
     info '  Already up-to-date';
   }
 
+  info 'Computing binary sketch size';
+  if (($built_something && $run_step->('size')) || $force->('size')) {
+    $builder->compute_binary_size();
+    # Not printing 'Success' here because the command already has an output.
+  } else {
+    info '  No new binary built';
+  }
+
+  info 'Success!';
+}
+
+sub discover {
+  my ($config) = @_;
+
+  info 'Running board discovery. Be sure to run with "-l debug" to see the result...';
+  my @ports = App::ArduinoBuilder::Discovery::discover($config);
+  $config->set('builder.internal.ports' => \@ports);
+  info 'Success!';
+}
+
+sub select_port {
+  my ($config) = @_;
+
+  {
+    my $port = $config->get('builder.internal.selected_port', default => undef);
+    return $port if defined $port;
+  }
+
+  my @ports = @{$config->get('builder.internal.ports')};
+  # TODO: implement an exact match selection and an interactive selection.
+  fatal "You must pass the --target-port option to select the upload target" unless $config->exists('builder.upload.port');
+  my @targets = map { fc } split(/\s*,\s*/, $config->get('builder.upload.port'));
+  my $port = first { my $port = $_; any { $port->get('upload.port.lc_label') eq $_ || $port->get('upload.port.lc_address') eq $_ } @targets } @ports;
+  unless (defined $port) {
+    error "None of the found ports match the specified ones";
+    error "Found the following ports: %s", join(', ', map { $_->get('upload.port.label') } @ports);
+    error "Specified ports: %s", join(', ', @targets);
+    fatal "None of the specified ports (%s) can be found, can your target be found by the 'discover' command?", join(', ', @targets);
+  }
+
+  $config->set('builder.internal.selected_port' => $port);
+  return $port;
+}
+
+sub upload {
+  my ($config) = @_;
+
+  info 'Uploading binary to the board...';
+
+  my $port = select_port($config);
+  my $protocol = $port->get('upload.port.protocol');
+  my $tool = $config->get("upload.tool.${protocol}", default => $config->get('upload.tool.default', default => $config->get('upload.tool')));
+  my $tool_config = $config->filter("tools.${tool}");
+
+  # TODO: add a way to set the verbose mode, in which case the upload.params.verbose
+  # property should be copied, instead of upload.params.quiet.
+  # Reference: https://arduino.github.io/arduino-cli/0.32/platform-specification/#verbose-parameter
+  $tool_config->set('upload.verbose' => $tool_config->get('upload.params.quiet'), allow_override => 1);
+
+  my $upload_config = $config->filter("upload.${protocol}")->prefix('upload');
+  $upload_config->merge($tool_config);
+  $upload_config->merge($port);
+  # Note: $config is in the recursive base in $upload_config.
+
+  # TODO: Before executing the command, some boards require that we manually
+  # reset them through their Serial port.
+  # See the code: https://github.com/arduino/arduino-cli/blob/ad9ddb882016c2af10e0db3785a46122bc9cfb1f/commands/upload/upload.go#L370
+  # And the doc: https://arduino.github.io/arduino-cli/0.32/platform-specification/#1200-bps-bootloader-reset
+
+  my $cmd = $upload_config->get('upload.pattern');
+  debug "Upload configuration:\n%s", sub { $upload_config->dump('  ') };
+  default_runner()->run_forked(sub {
+        close STDIN;
+        execute_cmd($cmd);
+      });
+
+  info 'Success!';
+
+}
+
+sub monitor {
+  my ($config) = @_;
+
+  info 'Running board monitor. Press ctrl+c to exit...';
+
+  my $port = select_port($config);
+  App::ArduinoBuilder::Monitor::monitor($config, $port);
+
   info 'Success!';
 }
 
