@@ -4,7 +4,7 @@ use strict;
 use warnings;
 use utf8;
 
-use App::ArduinoBuilder::Logger;
+use App::ArduinoBuilder::Logger '!dump';
 use Exporter 'import';
 
 our @EXPORT_OK = qw(get_os_name);
@@ -14,7 +14,7 @@ our @EXPORT_OK = qw(get_os_name);
 
 sub new {
   my ($class, %options) = @_;
-  my $this = bless {config => {}, files => 0, options => {}}, $class;
+  my $this = bless {config => {}, files => 0, options => {%options}}, $class;
   $this->read_file($options{file}, %options) if $options{file};
   for my $f (@{$options{files}}) {
     $this->read_file($f, %options, no_resolve => 1);
@@ -38,6 +38,24 @@ sub read_file {
   return 1;
 }
 
+# Parses a Perl data-structure into this objet.
+# $data should be an hash-ref.
+sub parse_perl {
+  my ($this, $data, %options) = @_;
+
+  fatal 'Can’t parse non hash-ref data: "%s"', ref($data) unless ref($data) eq 'HASH';
+  while (my ($k, $v) = each %{$data}) {
+    my $key = exists $options{prefix} ? $options{prefix}.'.'.$k : $k;
+    if (ref $v) {
+      $this->parse_perl($v, %options, prefix => $key);
+    } else {
+      $this->set($key => $v, %options);
+    }
+  }
+
+  return 1;
+}
+
 sub size {
   my ($this) = @_;
   return scalar keys %{$this->{config}};
@@ -59,6 +77,7 @@ sub get {
   return $options{default} if !$this->exists($key) && exists $options{default};
   $options{allow_partial} = 1 if $options{no_resolve};
   my $v = _resolve_key($key, $this->{config}, %options, allow_partial => 1);
+  return $v if $options{allow_partial};
   fatal "Key '$key' does not exist in the configuration." unless defined $v;
   fatal "Key '$key' has unresolved reference to value '$1'." if $v =~ m/\{([^}]+)\}/ && !$options{allow_partial};
   return $v;
@@ -86,9 +105,9 @@ sub set {
 }
 
 sub append {
-  my ($this, $key, $value) = @_;
+  my ($this, $key, $value, $sep) = @_;
   if ($this->{config}{$key}) {
-    $this->{config}{$key} .= ' '.$value;
+    $this->{config}{$key} .= ($sep // ' ').$value;
   } else {
     $this->{config}{$key} = $value;
   }
@@ -99,7 +118,7 @@ sub _resolve_key {
   my ($key, $config, %options) = @_;
   return $options{with}{$key} if exists $options{with}{$key};
   if (not exists $config->{$key}) {
-    return $options{base}->get($key, %options{grep { $_ ne 'base'} CORE::keys %options}) if exists $options{base} && $options{base}->exists($key);
+    return $options{base}->get($key, %options{grep { $_ ne 'base'} CORE::keys %options}) if exists $options{base};
     fatal "Can’t resolve key '${key}' in the configuration." unless $options{allow_partial};
     return;
   }
@@ -161,6 +180,16 @@ sub filter {
   return $filtered;
 }
 
+sub prefix {
+  my ($this, $prefix) = @_;
+  my $prefixed = App::ArduinoBuilder::Config->new();
+  $prefixed->{options}{base} = $this;
+  while (my ($k, $v) = each %{$this->{config}}) {
+    $prefixed->{config}{$prefix.'.'.$k} = $v;
+  }
+  return $prefixed;
+}
+
 sub dump {
   my ($this, $prefix) = @_;
   my $out = '';
